/*----------------------------------------------------------------------
  File    : JNICoCo.c
  Contents: COntinuous time ClOsed Neuron Assembly Detection for Java
  Author  : Christian Borgelt
  History : 2013.11.21 file created
            2013.11.25 first version completed
            2013.12.04 pattern spectrum generation function added
            2013.12.11 possibility to abort execution added
            2013.12.12 bug in isr_toJObj() fixed (item mapping)
            2014.03.06 estimation of a pattern spectrum added
            2014.05.09 surrogate data via block permutations added
            2014.05.12 filtering border added to coconad parameters
            2014.06.17 bug in function psp_toJObj() fixed (bounds)
            2014.07.16 evaluation of item sequence flag added
            2014.08.27 adapted to modified item set reporter interface
            2014.09.26 description of return values added
            2014.09.30 parameters added to patspec() and estpsp()
            2014.10.08 error handling in tns_fromJObj() corrected
            2014.10.09 functions repinit() and repterm() added
            2014.10.17 error handling/exception throwing improved
            2014.10.24 changed from LGPL license to MIT license
            2015.03.03 signal handling added (interrupt/ctrl-c)
            2015.08.07 ensured signal handler removal on error
            2015.08.10 memory leak in result fixed (data->{sets,supps})
            2015.08.13 functions patspec() renamed to genpsp()
            2015.08.13 adapted to generic function ccn_genpsp()
            2016.03.24 perfect extension flag replaced by mode
            2016.10.03 adapted to extensions of coconad implementation
            2016.10.21 adapted to interface changes of coconad impl.
            2016.10.23 closed/maximal repository coded in operation mode
            2016.11.18 coconad miner object and interface introduced
----------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <float.h>
#include <time.h>
#include <assert.h>
#ifndef TRN_SURR
#define TRN_SURR
#endif
#ifndef ISR_PATSPEC
#define ISR_PATSPEC
#endif
#ifndef ISR_CLOMAX
#define ISR_CLOMAX
#endif
#ifndef PSP_ESTIM
#define PSP_ESTIM
#endif
#ifndef PSP_TRAIN
#define PSP_TRAIN
#endif
#include "sigint.h"
#include "random.h"
#include "coconad.h"
#include "ccnpsp.h"
#include "JNICoCo.h"

/*----------------------------------------------------------------------
  Preprocessor Definitions
----------------------------------------------------------------------*/
#define BLKSIZE     1024        /* block size for arrays */

/* --- error handling --- */
#define X_MEM       0           /* OutOfMemoryError */
#define X_NULL      1           /* NullPointerException */
#define X_ARG       2           /* IllegalArgumentException */
#define X_CLASS     3           /* ClassNotFoundException */
#define X_ABORT     4           /* InterruptedException */
#define X_NONE      5           /* dummy */

#define ERR_RTN(e)      error(e, X_NONE,  NULL)
#define ERR_MEM(e)      error(e, X_MEM,   "not enough memory")
#define ERR_NULL(e)     error(e, X_NULL,  "trains")
#define ERR_ARG(e,s)    error(e, X_ARG,   s)
#define ERR_CLS(e,s)    error(e, X_CLASS, s)
#define ERR_ABORT(e)    error(e, X_ABORT, "user abort")

/*----------------------------------------------------------------------
  Type Definitions
----------------------------------------------------------------------*/
typedef struct {                /* --- item set report data --- */
  JNIEnv       *env;            /* JNI environment */
  jobjectArray res;             /* result for Java function */
  jsize        size;            /* size of frequent item set array */
  jsize        cnt;             /* number of frequent item sets */
  jintArray    *sets;           /* found frequent item sets */
  jint         *supps;          /* support values for item sets */
  int          err;             /* error flag */
} REPDATA;                      /* (item set report data) */

typedef struct {                /* --- Java environment data --- */
  JNIEnv       *env;            /* Java Native Interface environment */
  jint         done[1];         /* number of completed data sets */
  jintArray    ctrl;            /* Java control variables */
} JDATA;                        /* (Java environment data) */

/*----------------------------------------------------------------------
  Parameter Functions
----------------------------------------------------------------------*/

static int get_target (jint target)
{                               /* --- get target code */
  int s;                        /* buffer for target pattern type */
  switch (target & coconad_JNICoCo_TYPEMASK) {
    case coconad_JNICoCo_PERMUTATION: s = CCN_PERMUT;   break;
    case coconad_JNICoCo_SEQUENCE:    s = CCN_SEQUENCE; break;
    default:                          s = CCN_ITEMSET;  break;
  }                             /* get the target pattern type */
  if (target & coconad_JNICoCo_APPEND)
    s |= CCN_APPEND;            /* get the append flag if present */
  switch (target & coconad_JNICoCo_SUBMASK & ~coconad_JNICoCo_APPEND) {
    case coconad_JNICoCo_FREQUENT: return s|ISR_FREQUENT;
    case coconad_JNICoCo_CLOSED:   return s|ISR_CLOSED;
    case coconad_JNICoCo_MAXIMAL:  return s|ISR_MAXIMAL;
  }                             /* evaluate target pattern subtype */
  return -1;                    /* return an error code */
}  /* get_target() */

/*--------------------------------------------------------------------*/

static int get_coconad (jint algo, jint target)
{                               /* --- get CoCoNAD algorithm code */
  if (algo == coconad_JNICoCo_AUTO)
    algo = (target & (CCN_PERMUT|CCN_SEQUENCE))
         ? coconad_JNICoCo_FILTER : coconad_JNICoCo_CMREPO;
  switch (algo) {               /* evaluate the algorithm code */
    case coconad_JNICoCo_BASIC:  return CCN_BASIC;
    case coconad_JNICoCo_FILTER: return CCN_FILTER;
    case coconad_JNICoCo_REPO:   return CCN_REPO;
  }
  return -1;                    /* return an error code */
}  /* get_coconad() */

/*--------------------------------------------------------------------*/

static int get_surr (jint surr)
{                               /* --- get surrogate method code */
  switch (surr) {               /* evaluate the surrogate method code */
    case coconad_JNICoCo_IDENTITY: return CCN_IDENTITY;
    case coconad_JNICoCo_RANDOM:   return CCN_RANDOM;
    case coconad_JNICoCo_DITHER:   return CCN_DITHER;
    case coconad_JNICoCo_SHIFT:    return CCN_SHIFT;
    case coconad_JNICoCo_KERNEST:  return CCN_KERNEST;
    case coconad_JNICoCo_PERMUTE:  return CCN_PERMUTE;
    case coconad_JNICoCo_BLOCK:    return CCN_BLOCK;
  }
  return -1;                    /* return an error code */
}  /* get_surr() */

/*--------------------------------------------------------------------*/

static int get_rand (jint rand)
{                               /* --- get density function code */
  switch (rand) {               /* evaluate the density function code */
    case coconad_JNICoCo_RECT:   return RNG_RECT;
    case coconad_JNICoCo_TRIANG: return RNG_TRIANG;
    case coconad_JNICoCo_NORMAL: return RNG_NORMAL;
  }
  return -1;                    /* return an error code */
}  /* get_rand() */

/*----------------------------------------------------------------------
  Auxiliary Functions
----------------------------------------------------------------------*/

static void* error (JNIEnv *env, int type, const char *msg)
{                               /* --- throw an exception */
  jclass cls;                   /* class of the exception */
  char   *name;                 /* name of the exception */

  assert(env);                  /* check the function arguments */
  sig_remove();                 /* remove the signal handler */
  switch (type) {               /* evaluate the exception type */
    case X_MEM:   name = "java/lang/OutOfMemoryError";         break;
    case X_NULL:  name = "java/lang/NullPointerException";     break;
    case X_ARG:   name = "java/lang/IllegalArgumentException"; break;
    case X_CLASS: name = "java/lang/ClassNotFoundException";   break;
    case X_ABORT: name = "java/lang/InterruptedException";     break;
    default:      name = NULL;
  }                             /* get the name and then the class */
  if (type != X_NONE) {         /* if to raise an exception */
    cls = (*env)->FindClass(env, name);
    (*env)->ThrowNew(env, cls, (msg) ? msg : "");
  }                             /* create and throw exception */
  return NULL;                  /* throw the exception and return */
}  /* error() */

/*--------------------------------------------------------------------*/

static TRAINSET* tns_fromJObj (JNIEnv *env, jobjectArray trains)
{                               /* --- create a train set */
  int      e = 0;               /* error flag */
  IDENT    id;                  /* item identifier */
  jsize    i, k, n;             /* loop variables */
  jobject  o;                   /* to traverse the Java arrays */
  jdouble  *t;                  /* (spike) times of a train */
  ITEMBASE *ibase;              /* underlying item base */
  TRAINSET *trnset;             /* created train set */
  TRAIN    *trn;                /* to create the trains */

  assert(env && trains);        /* check the function argument */
  ibase = ib_create(IB_OBJNAMES, 0, ST_INTFN, (OBJFN*)0);
  if (!ibase) return ERR_MEM(env);  /* create an item base */
  n = (*env)->GetArrayLength(env, trains);
  for (id = 0; id < (IDENT)n; id++) /* add all item names */
    if (ib_add(ibase, &id) < 0) break;      /* (integers) */
  if (id < (IDENT)n) { ib_delete(ibase); return ERR_MEM(env); }
  trnset = tns_create(ibase);   /* create a (spike) train set */
  if (!trnset) { ib_delete(ibase); return ERR_MEM(env); }
  for (i = 0; i < n; i++) {     /* traverse the (spike) trains */
    o = (*env)->GetObjectArrayElement(env, trains, i);
    k = (*env)->GetArrayLength(env, o);
    t = (*env)->GetDoubleArrayElements(env, o, NULL);
    trn = trn_create((ITEM)i, t, (TID)k);
    if      (!trn) e = -1;     /* create a (spike) train object */
    else if (tns_add(trnset, trn) < 0) {
      e = -1; trn_delete(trn);}/* add created (spike) train to set */
    (*env)->ReleaseDoubleArrayElements(env, o, t, JNI_ABORT);
    (*env)->DeleteLocalRef(env, o);
    if (e) break;              /* release the Java array */
  }                            /* and check for an error */
  if (e) { tns_delete(trnset, 1); return ERR_MEM(env); }
  return trnset;               /* return the created train set */
}  /* tns_fromJObj() */

/*--------------------------------------------------------------------*/

static int isr_Jborder (JNIEnv *env, ISREPORT *rep, jintArray border)
{                               /* --- set reporter filtering border */
  jsize n;                      /* loop variables, border length */
  jint  *b;                     /* elements of the border array */

  assert(env && rep);           /* check the function arguments */
  if (!border) return 0;        /* abort if no border is given */
  n = (*env)->GetArrayLength(env, border);
  if (n <= 0)  return 0;        /* empty borders need no processing */
  b = (*env)->GetIntArrayElements(env, border, NULL);
  while (n > 0) { --n;          /* traverse the border elements */
    if (isr_setbdr(rep, (ITEM)n, (RSUPP)b[n]) < 0)
      break;                    /* set the support values and */
  }                             /* afterward release Java array */
  (*env)->ReleaseIntArrayElements(env, border, b, JNI_ABORT);
  return (n <= 0) ? 0 : -1;     /* return an error status */
}  /* isr_Jborder() */

/*--------------------------------------------------------------------*/

static void isr_toJObj (ISREPORT *rep, void *data)
{                               /* --- report an item set */
  REPDATA   *rd = data;         /* type the data pointer */
  jsize     n;                  /* (new) item set array size */
  ITEM      k, m;               /* loop variable, array size */
  void*     p;                  /* resized array */
  jint      *set;               /* item set elements */
  jintArray jset;               /* new Java item set */

  if (rd->err) return;          /* if there was an error, do nothing */
  n = rd->size;                 /* get the current array size */
  if (rd->cnt >= n) {           /* if the item set array is full */
    n += (n > BLKSIZE) ? n >> 1 : BLKSIZE;
    p = realloc(rd->sets,  (size_t)n *sizeof(jintArray));
    if (!p) { rd->err = -1; return; }
    rd->sets = p;               /* enlarge the item set array */
    p = realloc(rd->supps, (size_t)n *sizeof(jint));
    if (!p) { rd->err = -1; return; }
    rd->supps = p;              /* enlarge the support array */
    rd->size  = n;              /* set the new array size */
  }
  m    = (jsize)isr_cnt(rep);   /* create object for item set */
  jset = (*rd->env)->NewIntArray(rd->env, m);
  if (!jset) { rd->err = -1; return; }
  set  = (*rd->env)->GetIntArrayElements(rd->env, jset, NULL);
  for (k = 0; k < m; k++)       /* map identifiers to items */
    set[k] = (jint)(ptrdiff_t)isr_itemobj(rep, isr_itemx(rep, k));
  (*rd->env)->ReleaseIntArrayElements(rd->env, jset, set, 0);
  rd->sets [rd->cnt]   = jset;  /* store item set and its support */
  rd->supps[rd->cnt++] = (jint)isr_supp(rep);
}  /* isr_toJObj() */

/*--------------------------------------------------------------------*/

static jobjectArray psp_toJObj (JNIEnv *env, PATSPEC *psp,
                                double scale, int format)
{                               /* --- report pattern spectrum */
  ITEM         z;               /* loop variable for sizes */
  SUPP         s;               /* loop variable for support values */
  size_t       n, k;            /* buffer for number of occurrences */
  jsize        i;               /* array index */
  jclass       cls;             /* class of object/integer arrays */
  jmethodID    cid;             /* id of constructor method */
  jintArray    jsizes, jsupps;  /* sizes and support values */
  jint         *sizes, *supps;  /* ditto, array elements */
  jdoubleArray jfrqs;           /* occcurence frequencies */
  jdouble      *frqs;           /* ditto, array elements */
  jobjectArray jpsp;            /* pattern spectrum as Java object */
  jobject      jpse;            /* pattern spectrum element */

  assert(env && psp);           /* check the function arguments */
  n = psp_sigcnt(psp);          /* get the number of signatures */
  if ((format == '=')           /* if list of objects (triplets) */
  ||  (format == '-')) {        /* as objects PatSpecElem */
    cls  = (*env)->FindClass(env, "java/lang/Object");
    jpsp = (*env)->NewObjectArray(env, (jsize)n, cls, NULL);
    if (!jpsp) return ERR_MEM(env); /* create a PatSpecElem array */
    cls  = (*env)->FindClass (env, "fim/PatSpecElem");
    if (!cls)  return ERR_CLS(env, "fim/PatSpecElem");
    cid  = (*env)->GetMethodID(env, cls, "<init>", "(IID)V");
    if (!cid)  return ERR_CLS(env, "fim/PatSpecElem");
    for (i = 0, z = psp_min(psp); z <= psp_max(psp); z++) {
      for (s = psp_min4sz(psp, z); s <= psp_max4sz(psp, z); s++) {
        k = psp_getfrq(psp,z,s);/* traverse the pattern signatures */
        if (k <= 0) continue;   /* skip empty counters */
        jpse = (*env)->NewObject(env, cls, cid, (jint)z, (jint)s,
                                 (jdouble)((double)k *scale));
        if (!jpse) return ERR_MEM(env);
        (*env)->SetObjectArrayElement(env, jpsp, i, jpse); i++;
      }                         /* store a pattern spectrum element */
    } }                         /* and advance the array index */
  else {                        /* if three arrays (size,supp,freq) */
    jsizes = (*env)->NewIntArray   (env, (jsize)n);
    jsupps = (*env)->NewIntArray   (env, (jsize)n);
    jfrqs  = (*env)->NewDoubleArray(env, (jsize)n);
    cls    = (*env)->FindClass(env, "java/lang/Object");
    jpsp   = (*env)->NewObjectArray(env, 3, cls, NULL);
    if (!jsizes || !jsupps || !jfrqs || !jpsp)
      return ERR_MEM(env);      /* create a Java pattern spectrum */
    (*env)->SetObjectArrayElement(env, jpsp, 0, jsizes);
    (*env)->SetObjectArrayElement(env, jpsp, 1, jsupps);
    (*env)->SetObjectArrayElement(env, jpsp, 2, jfrqs);
    sizes = (*env)->GetIntArrayElements   (env, jsizes, NULL);
    supps = (*env)->GetIntArrayElements   (env, jsupps, NULL);
    frqs  = (*env)->GetDoubleArrayElements(env, jfrqs,  NULL);
    for (i = 0, z = psp_min(psp); z <= psp_max(psp); z++) {
      for (s = psp_min4sz(psp, z); s <= psp_max4sz(psp, z); s++) {
        k = psp_getfrq(psp,z,s);/* traverse the pattern signatures */
        if (k <= 0) continue;   /* skip empty counters */
        sizes[i]   = (jint)z; supps[i] = (jint)s;
        frqs [i++] = (jdouble)((double)k *scale);
      }                         /* store the signature elements */
    }                           /* in the corresponding arrays */
    (*env)->ReleaseIntArrayElements   (env, jsizes, sizes, 0);
    (*env)->ReleaseIntArrayElements   (env, jsupps, supps, 0);
    (*env)->ReleaseDoubleArrayElements(env, jfrqs,  frqs,  0);
  }                             /* release the arrays */
  assert((size_t)i == n);       /* check signature counter */
  return jpsp;                  /* return created pattern spectrum */
}  /* psp_toJObj() */

/*--------------------------------------------------------------------*/

static int repinit (JNIEnv *env, REPDATA *data, ISREPORT *isrep,
                    jstring report)
{                               /* --- initialize reporting */
  const jchar *s;               /* report indicator string */
  jchar       c;                /* report indicator character */

  assert(env                    /* check the function arguments */
  &&     data && isrep && report);
  data->sets  = NULL;           /* clear reporting arrays */
  data->supps = NULL;
  data->err   = 0;              /* initialize the error indicator */
  s = (*env)->GetStringChars(env, report, NULL);
  c = s[0];                     /* get the report indicators */
  (*env)->ReleaseStringChars(env, report, s);
  if ((c == '#') || (c == '|')  /* if to get a pattern spectrum */
  ||  (c == '=') || (c == '-')) /* #,|: 3 columns, =,-: triplets */
    return isr_addpsp(isrep, NULL);
  data->env   = env;            /* note the JNI environment and */
  data->res   = NULL;           /* initialize the report structure */
  data->size  = data->cnt = 0;  /* and set the reporting function */
  isr_setrepo(isrep, isr_toJObj, data);
  return 0;                     /* return 'ok' */
}  /* repinit() */

/*--------------------------------------------------------------------*/

static int repterm (JNIEnv *env, REPDATA *data, ISREPORT *isrep,
                    jstring report)
{                               /* --- terminate reporting */
  jsize        i;               /* loop variable */
  const jchar  *s;              /* report indicator string */
  jchar        c;               /* report indicator character */
  jclass       cls;             /* class of integer arrays */
  jobjectArray jsets;           /* found item sets as Java array */
  jintArray    jsupps;          /* support values as Java array */

  assert(env                    /* check the function arguments */
  &&     data && isrep && report);
  if (data->err) return data->err = -1;
  s = (*env)->GetStringChars(env, report, NULL);
  c = s[0];                     /* get the report indicators */
  (*env)->ReleaseStringChars(env, report, s);
  if ((c == '#') || (c == '|')  /* if to get a pattern spectrum */
  ||  (c == '=') || (c == '-')){/* #,|: 3 columns, =,-: triplets */
    data->res = psp_toJObj(env, isr_getpsp(isrep), 1.0, c);
    return data->err = (data->res) ? 0 : 1;
  }                             /* make Java pattern spectrum */
  if (data->err != 0) return data->err;
  cls    = (*env)->FindClass(env, "[I");
  jsets  = (*env)->NewObjectArray(env, data->cnt, cls, NULL);
  jsupps = (*env)->NewIntArray(env, data->cnt);
  cls    = (*env)->FindClass(env, "java/lang/Object");
  data->res = (*env)->NewObjectArray(env, 2, cls, NULL);
  if (!jsets || !jsupps || !data->res) return data->err = -1;
  for (i = 0; i < data->cnt; i++)  /* create and fill result arrays */
    (*env)->SetObjectArrayElement(env, jsets, i, data->sets[i]);
  (*env)->SetIntArrayRegion(env, jsupps, 0, data->cnt, data->supps);
  (*env)->SetObjectArrayElement(env, data->res, 0, jsets);
  (*env)->SetObjectArrayElement(env, data->res, 1, jsupps);
  return data->err = 0;         /* return the error status */
}  /* repterm() */

/*--------------------------------------------------------------------*/

static void repclean (REPDATA *data)
{                               /* --- clean up reporting data */
  if (data->sets)  { free(data->sets);  data->sets  = NULL; }
  if (data->supps) { free(data->supps); data->supps = NULL; }
}  /* repclean() */

/*--------------------------------------------------------------------*/

static void repfn (long cnt, void *data)
{                               /* --- progress reporting function */
  JDATA  *jd = data;            /* Java environment data */
  static long buf = 0;          /* already reported progress */

  jd->done[0] = (jint)buf;      /* set the control variable */
  (*jd->env)->SetIntArrayRegion(jd->env, jd->ctrl, 0, 1, jd->done);
  if ((cnt != buf) && ((cnt % 20) == 0))
    fprintf(stderr, "%10ld\b\b\b\b\b\b\b\b\b\b", buf = cnt);
}  /* repfn() */                /* report number of data sets */

/*--------------------------------------------------------------------*/
/* package : coconad                                                  */
/* class   : JNICoCo                                                  */
/* function: coconad (static)                                         */
/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_coconad_JNICoCo_coconad
  (JNIEnv *env, jclass objc, jobjectArray trains,
   jint target, jdouble width, jint supp, jint zmin, jint zmax,
   jstring report, jint algo, jint jmode, jintArray border)
{                               /* --- run the CoCoNAD algorithm */
  int      r;                   /* result of function call */
  TRAINSET *trnset;             /* set of (spike) trains */
  ITEM     n;                   /* number of (spike) trains */
  ISREPORT *isrep;              /* item set reporter */
  COCONAD  *coconad;            /* coconad miner */
  REPDATA  data;                /* data for item set reporting */
  int      mode = CCN_DEFAULT;  /* processing mode */
  int      mrep = 0;            /* reporting mode */

  assert(env);                  /* check the function arguments */
  if (!trains) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target);  /* translate the target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if (width  < 0) return ERR_ARG(env, "width must be >= 0");
  if (supp   < 0) supp = -supp; /* ensure non-negative support */
  if (zmin   < 1) return ERR_ARG(env, "zmin must be > 0");
  if (zmax   < 1) zmax = ITEM_MAX;  /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  algo = get_coconad(algo, target); /* translate algorithm code */
  if (algo   < 0) return ERR_ARG(env, "invalid algorithm variant");
  if (algo != CCN_REPO) mrep |= ISR_NOFILTER;
  if (jmode & coconad_JNICoCo_NOPERFECT) mode &= ~CCN_PERFECT;
  if (jmode & coconad_JNICoCo_NOSORT)    mode &= ~CCN_REORDER;
  if (jmode & coconad_JNICoCo_REPOLIST)  mode |=  CCN_REPOLIST;
  if (jmode & coconad_JNICoCo_REPOTREE)  mode |=  CCN_REPOTREE;
  if (jmode & coconad_JNICoCo_REPOCHAIN) mode |=  CCN_REPOCHAIN;

  /* --- create train set --- */
  sig_install();                /* install the signal handler */
  trnset = tns_fromJObj(env, trains);
  if (!trnset) return ERR_RTN(env); /* create and init. a train set */
  if (zmin > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax > ITEM_MAX) zmax = ITEM_MAX;
  coconad = ccn_create(target, width, supp,
  #ifdef CCN_OVERLAP            /* if influence map overlap support */
                       mix,     /* add. support mixing parameter */
  #endif
                       (ITEM)zmin, (ITEM)zmax, algo, mode);
  if (!coconad) {               /* create a coconad miner */
    tns_delete(trnset,  1); return ERR_MEM(env); }
  if (ccn_data(coconad, trnset, 0) != 0) {
    ccn_delete(coconad, 1); return ERR_MEM(env); }

  /* --- create item set reporter --- */
  n = tns_cnt(trnset);          /* get the number of trains */
  #ifndef CCN_OVERLAP           /* if standard (MIS) support */
  if (target & CCN_SEQUENCE)    /* if sequences (with poss. reps.) */
    n = tns_maxseq(trnset, width) +1;
  #endif                        /* find max. possible sequence len. */
  isrep = isr_createx(tns_base(trnset), n);    /* create a reporter */
  if (!isrep) { ccn_delete(coconad, 1); return ERR_MEM(env); }
  r = ccn_report(coconad, isrep);           /* prepare the reporter */
  if (r)      { ccn_delete(coconad, 1); return ERR_MEM(env); }
  if ((isr_Jborder(env, isrep, border)    != 0)
  ||  (repinit(env, &data, isrep, report) != 0)
  ||  (isr_setup(isrep) < 0)) { /* set up the item set reporter */
    ccn_delete(coconad, 1); return ERR_MEM(env); }

  /* --- frequent item set mining --- */
  r = ccn_mine(coconad);
  if (r >= 0) r = repterm(env, &data, isrep, report);
  repclean(&data);              /* get result and clean up */

  /* --- clean up --- */
  ccn_delete(coconad, 1);       /* delete the coconad miner */
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  if (r < 0) return ERR_MEM(env);
  if (r > 0) return ERR_RTN(env);
  sig_remove();                 /* remove the signal handler */
  return data.res;              /* return the created Java object */
}  /* Java_coconad_JNICoCo_coconad() */

/*--------------------------------------------------------------------*/
/* package : coconad                                                  */
/* class   : JNICoCo                                                  */
/* function: genpsp (static)                                          */
/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_coconad_JNICoCo_genpsp
  (JNIEnv *env, jclass objc, jobjectArray trains,
   jint target, jdouble width, jint supp, jint zmin, jint zmax,
   jstring report, jint algo, jint jmode, jint cnt,
   jdouble beg, jdouble end, jint surr, jint rand, jdouble sigma,
   jdouble delta, jint seed, jint cpus, jintArray ctrl)
{                               /* --- generate a pattern spectrum */
  int          mode = CCN_DEFAULT; /* search mode */
  PATSPEC      *psp = NULL;     /* created pattern spectrum */
  jobjectArray jpsp = NULL;     /* created Java pattern spectrum */
  TRAINSET     *trnset;         /* created train set */
  const jchar  *fmt;            /* report format/pat. spectrum flag */
  JDATA        jd;              /* Java environment data */

  assert(env);                  /* check the function arguments */
  if (!trains) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target);  /* translate the target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if (width  < 0) return ERR_ARG(env, "width must be >= 0");
  if (supp   < 0) supp = -supp; /* ensure non-negative support */
  if (zmin   < 1) return ERR_ARG(env, "zmin must be > 0");
  if (zmax   < 1) zmax = ITEM_MAX;  /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  algo = get_coconad(algo, target); /* translate algorithm code */
  if (algo   < 0) return ERR_ARG(env, "invalid algorithm variant");
  if (jmode & coconad_JNICoCo_NOPERFECT) mode &= ~CCN_PERFECT;
  if (jmode & coconad_JNICoCo_NOSORT)    mode &= ~CCN_REORDER;
  if (jmode & coconad_JNICoCo_REPOLIST)  mode |=  CCN_REPOLIST;
  if (jmode & coconad_JNICoCo_REPOTREE)  mode |=  CCN_REPOTREE;
  if (jmode & coconad_JNICoCo_REPOCHAIN) mode |=  CCN_REPOCHAIN;
  if (cnt   <= 0) cnt = 1;      /* check the number of data sets */
  surr   = get_surr(surr);      /* translate the surrogate code */
  if (surr  <  0) return ERR_ARG(env, "invalid surrogate method");
  if (surr  <= 0) cnt = 1;      /* adapt the number of data sets */
  rand   = get_rand(rand);      /* translate density function code */
  if (rand  <  0) return ERR_ARG(env, "invalid density function");
  if (sigma <  0) return ERR_ARG(env, "sigma must not be negative");
  if (delta <= 0) return ERR_ARG(env, "delta must be positive");

  /* --- generate pattern spectrum --- */
  sig_install();                /* install the signal handler */
  trnset = tns_fromJObj(env, trains);
  if (!trnset) return ERR_RTN(env); /* create and init. a train set */
  if (zmin > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax > ITEM_MAX) zmax = ITEM_MAX;
  jd.env     = env;             /* initialize Java environment data */
  jd.done[0] = 0; jd.ctrl = ctrl;
  psp = ccn_genpsp(trnset, target, width, supp, (ITEM)zmin, (ITEM)zmax,
                   algo, mode, (size_t)cnt, beg, end, surr, rand,
                   sigma, delta, (long)seed, (int)cpus, repfn, &jd);
  if (psp) {                    /* estimate a pattern spectrum */
    fmt = (*env)->GetStringChars(env, report, NULL);
    jpsp = psp_toJObj(env, psp, 1.0/(double)cnt, fmt[0]);
    (*env)->ReleaseStringChars(env, report, fmt);
    psp_delete(psp);            /* turn it into a Java object and */
  }                             /* delete the pattern spectrum */
  tns_delete(trnset, 1);        /* delete the train set */
  if (!jpsp) return ERR_MEM(env);
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  sig_remove();                 /* remove the signal handler */
  return jpsp;                  /* return the created Java object */
}  /* Java_coconad_JNICoCo_genpsp() */

/*--------------------------------------------------------------------*/
/* package : coconad                                                  */
/* class   : JNICoCo                                                  */
/* function: estpsp (static)                                          */
/*--------------------------------------------------------------------*/

JNIEXPORT jobjectArray JNICALL Java_coconad_JNICoCo_estpsp
  (JNIEnv *env, jclass objc, jobjectArray trains,
   jint target, jdouble width, jint supp, jint zmin, jint zmax,
   jstring report, jint equiv, jdouble alpha, jint smpls, jint seed)
{                               /* --- estimate a pattern spectrum */
  PATSPEC      *psp = NULL;     /* created pattern spectrum */
  jobjectArray jpsp = NULL;     /* created Java pattern spectrum */
  TRAINSET     *trnset;         /* created train set */
  const jchar  *fmt;            /* to evaluate the report string */

  assert(env);                  /* check the function arguments */
  if (!trains) return ERR_NULL(env);

  /* --- evaluate function arguments --- */
  target = get_target(target);  /* translate the target code */
  if (target < 0) return ERR_ARG(env, "invalid target type");
  if (width  < 0) return ERR_ARG(env, "width must be >= 0");
  if (supp   < 0) supp = -supp;    /* ensure non-negative support */
  if (zmin   < 1) return ERR_ARG(env, "zmin must be > 0");
  if (zmax   < 1) zmax = ITEM_MAX; /* check the size range */
  if (zmax   < zmin) return ERR_ARG(env, "zmax must be >= zmin");
  if (equiv <= 0) equiv = 1;    /* check the number of data sets */
  if (smpls <= 0) return ERR_ARG(env, "smpls must be > 0");
  if (seed  == 0) seed = (jint)time(NULL);

  /* --- estimate pattern spectrum --- */
  sig_install();                /* install the signal handler */
  trnset = tns_fromJObj(env, trains);
  if (!trnset) return ERR_RTN(env);/* create a train set and */
  tns_sort(trnset, +1);         /* sort the trains by points */
  if (zmin > ITEM_MAX) zmin = ITEM_MAX;
  if (zmax > ITEM_MAX) zmax = ITEM_MAX;
  psp = ccn_estpsp(trnset, target, width, supp, (ITEM)zmin, (ITEM)zmax,
                   (size_t)equiv, alpha, (size_t)smpls, seed);
  if (psp) {                    /* estimate a pattern spectrum */
    fmt = (*env)->GetStringChars(env, report, NULL);
    jpsp = psp_toJObj(env, psp, 1.0/(double)equiv, fmt[0]);
    (*env)->ReleaseStringChars(env, report, fmt);
    psp_delete(psp);            /* turn it into a Java object and */
  }                             /* delete the pattern spectrum */
  tns_delete(trnset, 1);        /* delete the train set */
  if (!jpsp) return ERR_MEM(env);
  if (sig_aborted()) { sig_abort(0); return ERR_ABORT(env); }
  sig_remove();                 /* remove the signal handler */
  return jpsp;                  /* return the created Java object */
}  /* Java_coconad_JNICoCo_estpsp() */

/*--------------------------------------------------------------------*/
/* package : coconad                                                  */
/* class   : JNICoCo                                                  */
/* function: abort (static)                                           */
/*--------------------------------------------------------------------*/

JNIEXPORT void JNICALL Java_coconad_JNICoCo_abort
(JNIEnv *env, jclass objc, jint state)
{ sig_abort((int)state); }      /* --- set the abort state */
